#include "include/mns.h"

FILE *in_file;

struct servicetype *servlist;
int services = 0;

void scan(struct in_addr);
int load_services(struct servicetype **);

void sighandler(int signo) { /* dummy signal handler */ }

int vuln(char *scan_target, int children) {
   static char buffer[60], *host;
   static int i, status;

   SAin vuln;
   int ff, cn = 0;

   if(!(services = load_services(&servlist))) {
      puts((char *)servlist);
      exit(EXIT_FAILURE);
   }
   
   if(!(mns_opts & HOSTLIST)) {
      printf("Scanning single target '%s'\n", scan_target);
      if(ipv4_rslvhost(scan_target, &vuln)) scan(vuln.sin_addr);
      else printf(" * invalid target %s", scan_target);
      return 1;
   }
   /* open file */

   if ((in_file = fopen (scan_target, "r")) == NULL) {
      printf(" * error opening %s\n", scan_target);
      exit (EXIT_FAILURE);
   }

   printf("Reading hosts from file '%s'\n", scan_target);

   for (i = 0; i < children; i++) {
      fgets(buffer, 60, in_file);
      if(feof(in_file)) break;
      host = strtok(buffer, "\n ");
      if(host == NULL) continue;

      if(ipv4_rslvhost(host, &vuln)) {
         cn++;
         if (!(ff = fork ())) {
            scan (vuln.sin_addr);
	    exit (0);
         }
         else if (ff == -1)
            printf ("fork failed!\n");
      } else printf ("* no route to '%s'\n", host);
   }

   if (!feof (in_file)) {
      for (;;) {
         wait (&status);
         fgets(buffer, 60, in_file);
         if(feof(in_file)) break;
         host = strtok(buffer, "\n ");
         if(host == NULL) continue;

         if (ipv4_rslvhost (host, &vuln)) {
            cn++;
            if (!(ff = fork ())) {
               scan (vuln.sin_addr);
               exit (0);
            } else if (ff == -1) printf ("fork failed!\n");
         } else printf ("* no route to '%s'\n", host);
      }
   }

   fclose (in_file);

   while (!((waitpid (-1, &status, WNOHANG) == -1) && (errno == ECHILD)));

   return cn;
}

void scan (struct in_addr target) {
   struct in_addr localip;
   struct node *node1st, *walk, *ntemp;
   char *hostname;
   int *portlist = NULL;
   int i, pcount;

   if (mns_opts & RESOLVE) {
      hostname = malloc (60);
      ipv4_showname (AF_INET, &target, hostname, 60);
   } else {
      hostname = malloc (16);
      ipv4_ntop (AF_INET, &target, hostname, 16);
   }
   
   node1st = walk = calloc(1, sizeof(struct node));

   walk = stealthscan(target, walk, &portlist, &localip);
   if(mns_opts & UDPSCAN) walk = udpportscan(target, walk, localip);
   if(walk != node1st) walk = checkos(target, walk, portlist, localip);
   
   if(portlist != NULL) {
      pcount = portlist[0] + 1;
      
      for (i = 1; i < portlist[0] + 1; i++) {
         switch (portlist[i]) {
            case 111: if(mns_opts & VULNRPC ) walk = rpc (target, walk); break;
            case 80:  if(mns_opts & VULNCGI ) walk = cgi (target, walk); break;
         }
      }
      
      stralloc(walk, "\n\tOther Possible Vulns ->");
      ntemp = walk;
      
      for (i = 1; i < portlist[0] + 1; i++) {
         switch (portlist[i]) {
            case 21:  if(mns_opts & VULNFTP ) walk = ftp (target, walk); break;
            case 23:  if(mns_opts & WINGATE ) walk = wingate (target, walk); break;
            case 110: if(mns_opts & VULNQPOP) walk = qpop (target, walk); break;
            case 143: if(mns_opts & VULNIMAP) walk = imap (target, walk); break;
            case 6000:if(mns_opts & XOPEN   ) walk = xopen(target, walk); break;
         }
      }
      
      if(walk == ntemp) stralloc(walk, "( none )");
   }

   if (node1st->next != NULL) {
      printf ("\nHost: %s", hostname);
      for (walk = node1st; walk->next != NULL; walk = walk->next)
         printf (" %s", walk->string);
   }
   putchar('\n');

   for (walk = node1st; walk != NULL;) {
      ntemp = walk->next;
      free (walk);
      walk = ntemp;
   }
}

int load_services(struct servicetype **st_ptr) {
   char buffer[60];
   char *token1 = NULL,
        *token2 = NULL;
   int i, j;

   if ((in_file = fopen (PORT_FILE, "r")) == NULL) {
      *st_ptr = malloc(sizeof(" * error opening " PORT_FILE));
      strcpy((char *)*st_ptr, " * error opening " PORT_FILE);
      return 0;
   }

   for(j = 0;;) {
      fgets(buffer, 60, in_file);
      if(feof(in_file)) break;
      if((token1 = strtok(buffer, " \t\n")) == NULL) continue;
      if(*token1 == '#') continue;
      if((token2 = strtok(NULL, " \t\n")) == NULL) continue;
      if(atoi(token2) > 0) j++;
   }

   if(!j) {
      *st_ptr = malloc(sizeof(" * " PORT_FILE
                              " contains no valid configuration data"));
      strcpy((char *)*st_ptr, " * " PORT_FILE
                              " contains no valid configuration data");
      return 0;
   }
   
   *st_ptr = malloc(j * sizeof(struct servicetype));
   
   fseek(in_file, 0, SEEK_SET);
   
   for(i = 0; i < j; i++) {
      for(;;) {
         fgets(buffer, 60, in_file);
         if((token1 = strtok(buffer, " \t\n")) == NULL) continue;
         if(*token1 == '#') continue;
         if((token2 = strtok(NULL, " \t\n")) == NULL) continue;
         if(atoi(token2) > 0) break;
      }
      
      (*st_ptr)[i].name = calloc(1, min(strlen(token1) + 1, 16));
      memcpy((*st_ptr)[i].name, token1, min(strlen(token1), 15));
      (*st_ptr)[i].port = atoi(token2);
   }
   
   fclose(in_file);
   return j;
}
