#include "include/mns.h"

#define ACK_RAND 200
#define PACKETS 6
#define PORT_CLOSED 1

#define TH_XXX 0x40
#define TH_YYY 0x80

#define MIN_PKT 200

#define SRC 0
#define DST 1

#define pcap_ip(a) ((struct ip *)(p + (a)))

/* local structures */
struct fingerprint {
   u_short set:2,
           seq:2,
           ackset:3,
           urg:1;
   u_long ack;
   u_short win;
   u_short flag;
};

typedef struct {
   u_long seq;
   u_long ack;
} hdrinfo;

/* function prototypes */
int init_packet_devices(void);
void term_packet_devices(void);
char *checkosmain(void);
void sendtcp(hdrinfo *, u_short);
int synscan(int *, hdrinfo);
int finscan(int *, hdrinfo);
struct in_addr get_local_ip(void);
u_long get_packet(struct ip *, char *, int, int);
void pcap_alarm(int sigio);
void sendudp(int);
int udpscan(int *);

/* global variables */

struct in_addr local_ip; /* local ip address */
struct in_addr target_ip; /* target ip address */

struct node **walk_global;

time_t timedelay;

u_short src_port; /* source port */
u_short dst_port; /* destination port */

int pcap_off; /* header offset for pcap_next captured packets */
int clicks;

pcap_t *pd_raw; /* packet capture descriptor */
int sd_tcp; /* tcp socket descriptor */
int sd_icmp; /* icmp socket descriptor */

struct node *stealthscan (struct in_addr target,
                          struct node *walk,
                          int **plist, struct in_addr *hostaddress)
{
   int (*scanfunc)(int *, hdrinfo);
   hdrinfo s_ack;
   struct tcphdr tcp;
   struct ip ip;
   int i, j, k;
   int *pschkd;
   time_t timeout;
   long ack_diff = 0;
   char buffer[80];

   srand(time(NULL));

   pschkd = (int *)calloc(services * sizeof(int), 1);

   walk_global = &walk;
   
   if(!init_packet_devices()) return walk;

   target_ip = target;

   local_ip = get_local_ip();
   
   if(!local_ip.s_addr) {
      stralloc(walk, "\n\t- no response from target machine");
      *plist = NULL;
      return walk;
   } else *hostaddress = local_ip;

   /* this is a TH_SYN | TH_FIN check that happens to be are really useful
    * check.  If an ACK of +2 is received then we know almost for sure
    * that we're looking at a Windows based system.
    * If we get no response, we either hit an open port on a *NIX based
    * machine or the machine is firewalled.
    * If we get ACK +2 or hit a firewalled machine, FIN scan won't work.
    * As far as I know, I'm the first one who has come up with this test.
    */
    
   s_ack.ack = 0;
   src_port = htons((rand() % 10000) + 5000);
   s_ack.seq = rand();
   
   sendtcp(&s_ack, TH_SYN | TH_FIN);

   for(i = j = 0; i < 2; i++) { /* we try twice just to make sure */
      timeout = time(NULL); /* we also calibrate the time delay */
      while(timeout + TIMEOUT > time(NULL))
         if(get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC))
            if(src_port == tcp.th_dport) { /* response to us? */
               if(ntohl(tcp.th_ack)) /* did we get an ACK? */
                  ack_diff = ntohl(tcp.th_ack) - s_ack.seq; /* ACK-SEQ */
               j = 1; /* yes, we got a valid packet */
               break;
            }
      if(j) break;
   }
   
   timedelay = time(NULL) - timeout + 2; /* calibration for portscan */

   if ((ack_diff == 2) || (!j)) { /* is our response weird? */
      stralloc(walk, "\n\t- non-standard response, reverting to SYN scan");
      stralloc(walk, "\n\t- target likely to be a Windows based machine");
      scanfunc = synscan; /* weird tagets get SYN scaned -- more stable */
   } else scanfunc = finscan; /* go for the FIN cause our target is ok */

   if(mns_opts & SYNSCAN) scanfunc = synscan; /* forced SYN scan? */

   /* if we get too few open ports using a FIN scan, our target is suspect
    * of being extra-weird, so let's SYN them. examples of this are IRIX
    * and HP-UX but there are a few others
    */

   if((j = scanfunc(pschkd, s_ack)) < 2)
      if(scanfunc == finscan) {
         stralloc(walk, "\n\t- FIN scan may have failed, trying SYN scan");
         j = synscan(pschkd, s_ack);
      }

   stralloc (walk, "\n\tStandard TCP service ports open ->");
   if(j) {

      *plist = malloc((j + 1) * sizeof(int));
      (*plist)[0] = j;

      for (i = k = 0; i < services; i++)
	{
	  if (!pschkd[i]) {
	      k++;
	      (*plist)[k] = servlist[i].port;
	      sprintf (buffer, "\n\t\t%3i  tcp\t%s",
		       servlist[i].port, servlist[i].name);
	      stralloc (walk, buffer);
          }
	}
    }
  else {
     stralloc (walk, "( none )");
     *plist = NULL;
  }

  term_packet_devices();

  return walk;
}

struct node *checkos (struct in_addr target,
                      struct node *walk,
                      int *plist,
                      struct in_addr hostaddress)
{
   char *buffer = NULL;
   int i, pcount;
   char ham_asc[2];

   if (plist != NULL) pcount = plist[0] + 1;
   else return walk;

   walk_global = &walk;

   if(!init_packet_devices()) return walk;
   local_ip = hostaddress;

   for (i = 1; i < min(pcount, 3); i++) {
      dst_port = htons(plist[i]);
      buffer = checkosmain();
      if (buffer != NULL) {
         ham_asc[0] = buffer[0];
	 ham_asc[1] = '\0';
	 stralloc (walk, "\n\tTCP Sequence Prediction Difficulty ->");
         stralloc (walk, ham_asc);
	 switch (ham_asc[0]) {
            case '0':
               stralloc (walk, "( indeterminate )");
	       break;
            case '1':
	       stralloc (walk, "( easy )");
	       break;
	    case '2':
	       stralloc (walk, "( easy )");
               break;
	    case '3':
	       stralloc (walk, "( moderate )");
	       break;
	    case '4':
	       stralloc (walk, "( moderate )");
	       break;
	    case '5':
	       stralloc (walk, "( hard )");
	       break;
	    case '6':
	       stralloc (walk, "( infeaseable )");
	       break;
	    case '7':
	       stralloc (walk, "( impossible )");
	       break;
	    case '8':
	       stralloc (walk, "( impossible )");
	       break;
	    default:
	       stralloc (walk, "( hamming error )");
	       break;
	    }
	    stralloc (walk, "\n\tOperating System ->");
	    stralloc (walk, &buffer[1]);
	    break;
         }
         if (buffer != NULL)
	   break;
      }

   if (buffer != NULL) free (buffer);
   else stralloc (walk, "\n\tOperating System -> ( indeterminate )");

  term_packet_devices();

   return walk;
}

char *checkosmain(void) {
  FILE *fs_cfg;
  char *osname;
  struct fingerprint pak[PACKETS];
  struct tcphdr tcp;
  struct ip ip;
  hdrinfo s_ack;
  u_short start, s;
  u_long myseq[PACKETS];
  u_long seq_mask = 0, last_seq = 0;
  time_t timeout;
  char line[60];
  int n, i, ham;

  for (n = 0; n < PACKETS; n++)
    bzero (&pak[n], sizeof (struct fingerprint));

  srand(time (NULL) & 0x0000ffff);
  start = s = (rand() % 5000) + 5000;
  s_ack.ack = 0;

#define sendtcppacket(a, b)\
   {s_ack.seq = myseq[a] = rand();\
    src_port = htons(s + a);\
    sendtcp(&s_ack, b);\
    usleep(10);}

  sendtcppacket (0, TH_SYN);
  sendtcppacket (1, TH_SYN | TH_ACK);
  sendtcppacket (2, TH_FIN | TH_ACK);
  sendtcppacket (3, TH_SYN | TH_FIN);
  sendtcppacket (4, TH_SYN | TH_XXX | TH_YYY);
  sendtcppacket (5, TH_SYN);	/* this extra syn is used for hamming distances
				 * an even number of responses are needed */
  timeout = time (NULL) + TIMEOUT;
  n = i = 0;

  do {
     if (get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC)) {
        if (dst_port == tcp.th_sport) {
           if (ntohs(tcp.th_dport) - start < PACKETS) {
              n = ntohs(tcp.th_dport) - start;
              if ((n < 0) || (n >= PACKETS)) continue;
              if (pak[n].set) continue;
              if (!n && !ntohs(tcp.th_win)) return NULL;
              pak[n].seq = tcp.th_seq != 0;
              if (pak[n].seq) {
                 seq_mask ^= tcp.th_seq;
                 last_seq = tcp.th_seq;
                 i++;
              }
              pak[n].ack = ntohl (tcp.th_ack);
              pak[n].ackset = tcp.th_ack != 0;
              if (pak[n].ackset && (myseq[n] > pak[n].ack))
                 pak[n].ackset = 2;
              if (pak[n].ackset)
                 pak[n].ack = (myseq[n] < pak[n].ack) ?
                 (pak[n].ack - myseq[n]) :
                 (myseq[n] - pak[n].ack);
              if (pak[n].ack > ACK_RAND) {
                 pak[n].ackset = 3;
                 pak[n].ack = 0;
              }
              pak[n].win = ntohs (tcp.th_win);
              pak[n].flag = tcp.th_flags;
	      pak[n].set = 1;
              pak[n].urg = tcp.th_urp != 0;
           }
        }
     }
     usleep (10);
  }
  while ((timeout > time (NULL)) && (n != PACKETS - 1));
  
  /* calculate hamming distance */
  if (i % 2) seq_mask ^= last_seq;
  ham = hamming (seq_mask);
  if (ham > 8) ham = 16 - ham;

  /* parse fingerprints */
  if ((fs_cfg = fopen (OSCFGFILE, "r")) != NULL)
    {
      struct fingerprint parse;
      int flag1 = 0, found = 0, total = 0;
      char *token;

      osname = malloc (60);

      while (fgets (line, 60, fs_cfg))
	{
	  if (line[0] == '#')
	    continue;		/* COMMENT LINE */
	  if (line[0] == '\n')
	    {			/* BLANK LINE */
	      if (flag1 && (found == total))
		{
		  fclose(fs_cfg);
		  osname[0] = (char)ham + '0';
		  return osname;
		};
	    }
	  if (line[0] == '*')
	    {			/* OS FINGERPRINT HEADER */
	      if (strlen (line) > 3)
		{
		  strcpy (osname, &line[1]);
		  osname[strlen (osname) - 1] = '\0';
		}
	      flag1 = 1;
	      total = found = 0;
	      continue;
	    }

	  /* line number */
	  token = strtok (line, " ");
	  if (isdigit (token[0]))
	    n = atoi (token);
	  else
	    {
	      flag1 = total = found = 0;
	      continue;
	    }

/*-- seq --*/
	  if ((token = strtok (NULL, " ")) != NULL)
	    parse.seq = atoi (token);
	  parse.set = isdigit (token[0]) ? 1 : 0;

/*-- ack --*/
	  if ((token = strtok (NULL, " ")) != NULL)
	    {
	      switch (token[0])
		{
		case '+':
		  parse.ackset = 1;
		  parse.ack = atol (&token[1]);
		  break;
		case '-':
		  parse.ackset = 2;
		  parse.ack = atol (&token[1]);
		  break;
		case 'R':
		  parse.ackset = 3;
		  parse.ack = 0;
		  break;
		case 'N':
		default:
		  parse.ackset = 0;
		  parse.ack = 0;
		  break;
		}
	    }

/*-- win --*/
	  if ((token = strtok (NULL, " ")) != NULL)
	    parse.win = strtol (token, NULL, 16);

/*-- flags --*/
	  if ((token = strtok (NULL, " \n")) != NULL)
	    {
	      parse.flag = 0;
	      if (strchr (token, 'S'))
		parse.flag |= TH_SYN;
	      if (strchr (token, 'R'))
		parse.flag |= TH_RST;
	      if (strchr (token, 'A'))
		parse.flag |= TH_ACK;
	      if (strchr (token, 'F'))
		parse.flag |= TH_FIN;
	      if (strchr (token, 'P'))
		parse.flag |= TH_PUSH;
	      if (strchr (token, 'X'))
		parse.flag |= TH_XXX;
	      if (strchr (token, 'Y'))
		parse.flag |= TH_YYY;
	      if (strchr (token, 'U'))
		parse.urg = 1;
	      else
		parse.urg = 0;
	    }

	  total++;

	  /* test for no response */
	  if (!pak[n].set)
	    {
	      if (parse.set)
		total = found = flag1 = 0;
	      else
		found++;
	      continue;
	    }

	  if ((parse.ackset == pak[n].ackset) &&
	      (parse.ack == pak[n].ack) &&
	      (parse.seq == pak[n].seq) &&
	      (parse.win == pak[n].win) &&
	      (parse.flag == pak[n].flag) &&
	      (parse.urg == pak[n].urg))
	    {
	      found++;
	      continue;
	    }
	  else
	    {
	      total = found = flag1 = 0;
	      continue;
	    }
	}

      fseek(fs_cfg, 0L, SEEK_END);
      fclose(fs_cfg);
      if(osname != NULL) free(osname);
      return NULL;
    }

  puts("error: cannot open " OSCFGFILE);
  exit(EXIT_FAILURE);
}

/* Sends a TCP packet */
void sendtcp(hdrinfo *routing, u_short flags) {
  struct tcphdr tcp;
  SAin sa_dest;

  tcp.th_sport = src_port;
  tcp.th_dport = dst_port;
  tcp.th_seq = htonl (routing->seq);
  tcp.th_ack = 0;
  tcp.th_x2 = 0;
  tcp.th_off = 5;
  tcp.th_flags = flags;
  tcp.th_win = htons(0x1000 | (rand () % 0x1000));
  tcp.th_urp = 0;
  tcp.th_sum = 0;
/*-- TCP Checksum --*/
#ifdef SOLARIS_CKSUM_BUG
  tcp.th_sum = sizeof (struct tcphdr);
#else
  tcp.th_sum = tcp_cksum((char *)&tcp, sizeof(tcp), local_ip, target_ip);
#endif /* SOLARIS_CKSUM_BUG */

   sa_dest.sin_family = AF_INET;
   sa_dest.sin_addr = target_ip;
   sa_dest.sin_port = 0;
   
   if(sendto(sd_tcp, &tcp, sizeof(tcp), 0, (SA *)&sa_dest, sizeof(SAin)) < 0)
      perror("sending message");
}

u_long get_packet(struct ip *ip_ptr, char *packet, int pkt_len, int side) {
   struct ip ip2;
   struct pcap_pkthdr pcap_hdr;
   char *p = NULL;
   int i, max;

   signal(SIGALRM, &pcap_alarm);
   clicks = 0;

   if(ip_ptr == NULL) ip_ptr = &ip2;

   alarm(1);
   if((p = (char *)pcap_next(pd_raw, &pcap_hdr)) == NULL) return 0;
   alarm(0);

   if(!pcap_off) max = min(50, pcap_hdr.caplen);
   else max = pcap_off + 1;
   
   for(i = pcap_off; i < max; i++) {
      memcpy(ip_ptr, pcap_ip(i), sizeof(struct ip));

      ip_ptr->ip_sum = 0;
      ip_ptr->ip_sum = in_cksum((u_short *)ip_ptr, sizeof(struct ip));

      if(pcap_ip(i)->ip_sum == ip_ptr->ip_sum) {
         if( (side == SRC ?
             pcap_ip(i)->ip_src.s_addr :
             pcap_ip(i)->ip_dst.s_addr) == target_ip.s_addr) {

            bzero(packet, pkt_len);
            memcpy(packet, p + i + sizeof(struct ip),
                   min(pkt_len, pcap_hdr.caplen));
                   
            if(!pcap_off) pcap_off = i;
            return  (side == DST ?
             pcap_ip(i)->ip_src.s_addr :
             pcap_ip(i)->ip_dst.s_addr);
         } else break;
      }
   }
   
   return 0;
}

int init_packet_devices(void) {
   char errbuf[PCAP_ERRBUF_SIZE];

   pcap_off = 0;
   pd_raw = pcap_open_live(device, MIN_PKT << 2, 0, MIN_PKT, errbuf);
   if(pd_raw == NULL) {
      stralloc(*walk_global, "\n\t- opening pcap device:");
      stralloc(*walk_global, errbuf);
      return 0;
   }

   if((sd_tcp = socket(AF_INET, SOCK_RAW, IPPROTO_TCP)) < 0) {
      stralloc(*walk_global, "\n\t- opening raw socket (tcp)");
      stralloc(*walk_global, strerror(errno));
      return 0;
   }

   if((sd_icmp = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
      stralloc(*walk_global, "\n\t- opening raw socket (icmp)");
      stralloc(*walk_global, strerror(errno));
      return 0;
   }
   
   return 1;
}

void term_packet_devices(void) {
   close(sd_tcp);
   close(sd_icmp);
   pcap_close(pd_raw);
}

int synscan(int *pschkd, hdrinfo s_ack) {
  struct tcphdr tcp;
  struct ip ip;
  int i, j, found = 0;
  time_t timeout;

  for (j = 0; j < services; j++)
    pschkd[j] = PORT_CLOSED;

   for (i = 0; i < services; i++) {
      dst_port = htons(servlist[i].port);

      s_ack.seq = rand ();
      sendtcp(&s_ack, TH_SYN);

      if(get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC))
         for (j = 0; j < services; j++)
            if(servlist[j].port == ntohs(tcp.th_sport)) {
               if(ntohs(tcp.th_win)) {
                  pschkd[j] = 0;
                  found++;
		  break;
               }
            }
   }

   timeout = time (NULL) + timedelay;

   while (timeout > time (NULL)) {
      if (get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC)) {
         for (j = 0; j < services; j++)
            if (servlist[j].port == ntohs (tcp.th_sport)) {
               if (ntohs (tcp.th_win)) {
                  pschkd[j] = 0;
                  found++;
                  break;
               }
            }
      }
   }
   return found;
}

int finscan(int *pschkd, hdrinfo s_ack) {
   struct tcphdr tcp;
   struct ip ip;
   int i, j, found = services;
   time_t timeout;
   u_short tcpflags = 0;

   for (i = 0; i < services; i++) {
      dst_port = htons(servlist[i].port);
      
      s_ack.seq = rand();
      
      switch (i % 3) { /* we actually use NUL/FIN/PSH scanning */
	case 0: tcpflags = 0x0;     break; /* NULL */
	case 1: tcpflags = TH_FIN;  break; /* FIN  */
	case 2: tcpflags = TH_PUSH; break; /* PUSH */
      }

      sendtcp(&s_ack, tcpflags);

      if (get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC)) {
         for(j = 0; j < services; j++)
            if(servlist[j].port == ntohs(tcp.th_sport)) {
               pschkd[j] = PORT_CLOSED;
               found--;
               break;
            }
      }
   }
   timeout = time (NULL) + timedelay;

   while (timeout > time (NULL)) {
      if(get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC)) {
         for (j = 0; j < services; j++)
            if(servlist[j].port == ntohs(tcp.th_sport)) {
               pschkd[j] = PORT_CLOSED;
               found--;
               break;
            }
      }
   }

   return found;
}

struct in_addr get_local_ip(void) {
   SAin sa_dest;
   struct in_addr ip_addr;
   struct icmp_h icmp;
   struct ip ip;
   time_t timeout;

   ip_addr.s_addr = 0;

   bzero(&icmp, sizeof(icmp));
   icmp.icmp_type = ICMP_ECHO;
   icmp.icmp_un.echo.id = getpid();
   icmp.icmp_sum = in_cksum((u_short *)&icmp, sizeof(icmp));

   sa_dest.sin_family = AF_INET;
   sa_dest.sin_addr = target_ip;
   sa_dest.sin_port = htons((rand() % 1000) + 1000);
   
   sendto(sd_icmp, &icmp, sizeof(icmp), 0, (SA *)&sa_dest, sizeof(SAin));
   bzero(&icmp, sizeof(icmp));

   timeout = time(NULL) + TIMEOUT;
   while(timeout > time(NULL))
      if(get_packet(&ip, (char *)&icmp, sizeof(icmp), DST))
         if(icmp.icmp_type == ICMP_ECHO)
            if(icmp.icmp_un.echo.id == getpid()) {
               ip_addr = ip.ip_src;
               break;
            }

   return ip_addr;
}

void pcap_alarm(int sigio) {
   SAin sa_dest;
   struct icmp_h icmp;

   sa_dest.sin_family = AF_INET;
   sa_dest.sin_addr = target_ip;
   sa_dest.sin_port = htons((rand() % 1000) + 1000);

   bzero(&icmp, sizeof(icmp));
   icmp.icmp_type = ICMP_ECHO;
   icmp.icmp_un.echo.id = getpid();
   icmp.icmp_sum = in_cksum((u_short *)&icmp, sizeof(icmp));
   
   sendto(sd_icmp, &icmp, sizeof(icmp), 0, (SA *)&sa_dest, sizeof(SAin));
   
   if(++clicks == 5) {
      printf("cannot receive messages on the interface '%s'\n", device);
      exit(EXIT_FAILURE);
   }
   
   alarm(1);
}

#define udps_icmp ((struct icmp_h *)buffer)
#define udps_ip ((struct ip *)&buffer[sizeof(struct icmp_h)])
#define udps_udp ((struct udp_h *)&buffer[sizeof(struct icmp_h) +\
                                          sizeof(struct ip)])

int udpscan(int *pschkd) {
   struct ip ip;
   char buffer[sizeof(struct ip) +
               sizeof(struct udp_h) +
               sizeof(struct icmp_h)];
   int i, j, sd_udp, found = services;
   time_t timeout;

   if((sd_udp = socket(AF_INET, SOCK_RAW, IPPROTO_UDP)) < 0) {
      perror("\n\t- opening raw socket (udp)");
      exit(EXIT_FAILURE);
   }

   for (i = 0; i < services; i++) {
      dst_port = htons(servlist[i].port);
      
      sendudp(sd_udp);

      if(!get_packet(&ip, buffer, sizeof(buffer), SRC)) continue;
      if(udps_icmp->icmp_type != ICMP_DEST_UNREACH) continue;
      if(udps_icmp->icmp_code != ICMP_PORT_UNREACH) continue;
      for(j = 0; j < services; j++)
         if(servlist[j].port == ntohs(udps_udp->udp_dport)) {
            pschkd[j] = PORT_CLOSED;
            found--;
            break;
         }

   }
   
   timeout = time(NULL) + TIMEOUT;

   while (timeout > time (NULL))
      if (get_packet(&ip, buffer, sizeof(buffer), SRC))
         if(udps_icmp->icmp_type == ICMP_DEST_UNREACH)
            if(udps_icmp->icmp_code == ICMP_PORT_UNREACH)
               for(j = 0; j < services; j++)
                  if(servlist[j].port == ntohs(udps_udp->udp_dport)) {
                     pschkd[j] = PORT_CLOSED;
                     found--;
                     break;
                  }

   close(sd_udp);
   return found;
}

void sendudp(int sd_udp) {
  struct udp_h udp;
  SAin sa_dest;

   udp.udp_sport = 0;
   udp.udp_dport = dst_port;
   udp.udp_len = htons(8);
   udp.udp_sum = 0;

   sa_dest.sin_family = AF_INET;
   sa_dest.sin_addr = target_ip;
   sa_dest.sin_port = dst_port;
   
   if(sendto(sd_udp, &udp, sizeof(udp), 0, (SA *)&sa_dest, sizeof(SAin)) < 0)
      perror("sending message");
}

struct node *udpportscan(struct in_addr target,
                         struct node *walk,
                         struct in_addr hostaddress)
{
   int i, j, k;
   int *pschkd;
   char buffer[80];

   srand(time(NULL));

   pschkd = (int *)calloc(services * sizeof(int), 1);

   walk_global = &walk;
   
   if(!init_packet_devices()) return walk;

   target_ip = target;
   local_ip = hostaddress;

   j = udpscan(pschkd);
   if(services - j < 10) {
      j = 0;
      stralloc (walk, "\n\t- UDP scan may have failed,"
                      " no results will be shown");
   }
   
   stralloc (walk, "\n\tStandard UDP service ports open ->");
   if(j) {
      for (i = k = 0; i < services; i++) {
         if (!pschkd[i]) {
            k++;
            sprintf (buffer, "\n\t\t%3i  udp\t%s",
            servlist[i].port, servlist[i].name);
            stralloc (walk, buffer);
         }
      }
   } else stralloc (walk, "( none )");

  term_packet_devices();

  return walk;
}

